/*
 * ex1_2.cc
 *
 * Validation code of blow-up solution of
 * u' = u^2 - v,
 * v' = u^3/3,
 *
 * Space compactification: Quasi-parabolic compactification
 * **Notice** This code is specialized for the shockwave (not automatic verification).
 *
 * written  Jul. 11, 2017 Akitoshi Takayasu (with kv-0.4.41)
 *
 */

#include <kv/ode-maffine.hpp>
// #include <kv/ode-maffine0.hpp>
#include <kv/kraw-approx.hpp>
#include <kv/eig.hpp>
const int size_of_ode = 2;    // Size of ODEs: fixed
const int k = 1; // quasi-homogeneous vector field of the order k+1: fixed
const int alph[] = {1,2}; // {alpha_i}: fixed
const int beta[] = {2,1}; //{beta_i}: fixed
const int gamm = 2; // c = alpha_1*beta_1 = ... = alpha_n*beta_n: fixed

namespace ub = boost::numeric::ublas;
ub::matrix< kv::interval<double> > Y;
ub::vector< kv::interval<double> > p;

// Define tilde{f}(x): you must edit for each ODE.
struct Func_tilde {
    template <class T> ub::vector<T> operator() (const ub::vector<T>& x) {
        ub::vector<T> y(size_of_ode);
        y(0) = pow(x(0),2)-x(1);
        y(1) = pow(x(0),3)/3.;
        return y;
    }
};
// Dominant term p_d(x): you must edit for each ODE.
struct Func_pd {
    template <class T> ub::vector<T> operator() (const ub::vector<T>& x) {
        ub::vector<T> y(size_of_ode);
        y(0) = pow(x(0),2)-x(1);
        y(1) = pow(x(0),3)/3.;
        return y;
    }
};
// Generator of desingularized vector field
struct Func {
    template <class T> ub::vector<T> operator() (const ub::vector<T>& x, const T& t) {
        Func_tilde f;
        int n = size_of_ode;
        ub::vector<T> y(n), y_tmp(n);
        T tmp, kappa;
        y_tmp = f(x);
        for (int i = 0; i < n; i++) {
          tmp = 0;
          kappa = 0;
          for (int j = 0; j < n; j++) {
            if (j!=i) {
              tmp += pow(x(j),2*beta[j]-1)/alph[j]*y_tmp(j);// except pow(~,1)?
              kappa += pow(x(j),2*beta[j]);
            }
          }
          kappa = 0.5/gamm*(1-pow(x(i),2*beta[i]))+0.5*(2.*gamm-1)/gamm*kappa;
          y(i) = kappa*y_tmp(i)-alph[i]*x(i)*tmp;
        }
        return y;
    }
};
// Func for kraw-approx
struct Func_krawapp {
    template <class T> ub::vector<T> operator() (const ub::vector<T>& x) {
        Func_pd f;
        int n = size_of_ode;
        ub::vector<T> y(n), y_tmp(n);
        T tmp;
        y_tmp = f(x);
        for (int i = 0; i < n; i++) {
          tmp = 0;
          for (int j = 0; j < n; j++) {
            if (j!=i) {
              tmp += pow(x(j),2*beta[j]-1)/alph[j]*y_tmp(j);
            }
          }
          y(i) = (1-pow(x(i),2*beta[i]))*y_tmp(i)-alph[i]*x(i)*tmp;
        }
        return y;
    }
};
// ode_callback function:
// Integrate the time variable (t) and check the validated solution is in the Lyapunov domain.
namespace kv{
template <class T> struct ode_callback_sample : ode_callback<T> {
    interval<T> &result2;
    T eps;

    ode_callback_sample(interval<T>& result2, T eps) : result2(result2), eps(eps) {}

    virtual bool operator()(const interval<T>& start, const interval<T>& end, const ub::vector< interval<T> >& x_s, const ub::vector< interval<T> >& x_e, const ub::vector< psa< interval<T> > >& result) const {
        int i, s;
        ub::vector< interval<T> > xp;
        kv::interval<T> Lx;
        s = result.size();
        psa< interval<T> > px2c; // 1-p(x)^{2c}
        px2c = 1.;
        for (i = 0; i<s; i++) {
          px2c -= pow(result(i),2*beta[i]);
        }
        if (k==1) {
          result2 += eval(integrate((1.0-0.5*(2*gamm-1)/gamm*px2c)*px2c),end-start);
        } else {
          result2 += eval(integrate((1.0-0.5*(2*gamm-1)/gamm*px2c)*pow(px2c,k)),end-start);
        }
        // Stopping criterion
        xp = x_e-p;
        Lx = inner_prod(prod(Y,xp),xp);
        if (Lx < pow(eps,(kv::interval<T>)2)) return false;
        return true;
    }
};
};

int main()
{
    int n = size_of_ode;    // # of unknowns
    ub::vector< kv::interval<double> > x, xp, y0, x0;
    kv::interval<double> start, end, c1, cn_tmp, cn, Tmax, t_n, y, Lx, kappa;
    std::cout.precision(17);
    // Initial value: x(0)
    x0.resize(n);
    // kappa = 1+sqrt(1+4*inner_prod(y0,y0));
    // for (int i = 0; i < n; i++) {
    //   x0(i) = 2*y0(i)/kappa;
    // }
    // x0(0) = 0.0; x0(1) = -0.1;
    x0(0) = -0.1; x0(1) = -0.1;
    // x0(0) = -0.1; x0(1) = 0.01;// Failed
    // x0(0) = -0.1; x0(1) = 0.0;
    // x0(0) = 0.1; x0(1) = 0;
    // For check
    kappa = 1.;
    for (int i = 0; i < n; i++) {
      kappa -= pow(x0(i),2*beta[i]);
    }
    // y0: initial vector of y(t): you must edit for each ODE.
    y0.resize(n);
    for (int i = 0; i < n; i++) {
      y0(i) = x0(i)/pow(kappa,alph[i]);
    }
    std::cout << "y0: " << y0 << '\n';

    // Validation of the blow-up solution
    // 1. Solve the desingularized ode until some time and determine the critical point at inifity:
    start = 0;
    end = 100;
    while (true) {
      x = x0;
      // kv::odelong_maffine(Func(), x, start, end,c1,p,ub::identity_matrix< kv::interval<double> >(1));
      kv::odelong_maffine(Func(), x, start, end, kv::ode_param<double>().set_verbose(0));
      p.resize(n);
      if (kv::krawczyk_approx(Func_krawapp(), mid(x), p, 15, 0)) {
        break;
      }
      end *= 1.5;
    }
    // std::cout << "end_time: " << end << '\n';
    std::cout << "p: " << p << "\n";

    // 2. Search the maximum range of the Lyapunov domain:
    // Calculate C1
    Func f;
    ub::vector< kv::interval<double> > y1,y1_tmp,y2,radius;
    ub::matrix< kv::interval<double> > Dfx, Ax;
    ub::vector< kv::autodif< kv::interval<double> > > dy1, dy2;
    // Initial vector of auto diff.
    y1.resize(n);
    for (int i = 0; i < n; i++) {
      // y1(i) = kv::interval<double>::hull(x(i),p(i));
      y1(i) = p(i);
    }
    dy1 = kv::autodif< kv::interval<double> >::init(y1);
    dy2 = f(dy1,(kv::autodif< kv::interval<double> >) 0); // f(dy1) = [f(y1), df/dx[y1](Jacobian) (=Dfx)]
    kv::autodif< kv::interval<double> >::split(dy2, y2, Dfx);
    // std::cout << "Df[p]: " << mid(Dfx) << std::endl;
    // Compute eigenvalue of Df(p)
    // Y=Re(V^-H*V^-1) s.t. V^-1*Dfx*V = D (diagonal matrix)
    ub::vector< kv::complex< kv::interval<double> > > lambda;
    ub::matrix< kv::complex<double> > V, D, R, RH, Y_comp, I;
    kv::eig(mid(Dfx),V,D);
    std::cout << "approximate eigenvalue:" << '\n';
    for (int i = 0; i < n; i++) {
      std::cout << D(i,i) << '\n';
    }
    // std::cout << "D: " << D << std::endl;
    kv::invert_comp(V,R);
    // std::cout << "R: " << R << std::endl;
    RH.resize(n,n);
    for (int i = 0; i < n; i++) {
      for (int j = 0; j < n; j++) {
        RH(j,i) = kv::complex<double>(R(i,j).real(),-R(i,j).imag());
      }
    }
    // std::cout << "R^H: " << RH << std::endl;
    I.resize(n,n);
    for (int i = 0; i < n; i++) {
      I(i,i) = 1;
    }
    // std::cout << "I: " << I << std::endl;
    Y_comp = prod(RH,I);
    Y_comp = prod(Y_comp,R);
    // std::cout << "Y_comp: " << Y_comp << std::endl;
    Y.resize(n,n);
    for (int i = 0; i < n; i++) {
      for (int j = 0; j < n; j++) {
        Y(i,j) = Y_comp(i,j).real();
      }
    }
    // std::cout << "Y: " << Y << std::endl;
    kv::veig(Y,lambda);
    // std::cout << "lambda_Y: " << lambda <<  std::endl;
    using std::abs;
    c1 = abs(lambda(0).real());
    for (int i = 1; i < n; i++) {
      if (c1 > abs(lambda(i).real())) {
        c1 = abs(lambda(i).real());
      }
    }
    c1 = 1./c1;
    std::cout << "c_1: " << c1 << std::endl;

    // Calculate Cn by the delta inflation scheme
    double eps_tmp = 1e-10, eps;// Radius of N
    y1_tmp.resize(n);
    radius.resize(n);
    while (true) {
      //------------------------- A(x) by using autodiff -----------------------
      for (int i = 0; i < n; i++) {
        radius(i) = kv::interval<double>(-eps_tmp,eps_tmp);
      }
      y1_tmp = p + radius*sqrt(c1);
      // std::cout << "y1 = " << y1 << std::endl;
      dy1 = kv::autodif< kv::interval<double> >::init(y1_tmp);
      dy2 = f(dy1,(kv::autodif< kv::interval<double> >) 0); // f(dy1) = [f(y1), df/dx[y1](Jacobian) (=Dfx)]
      kv::autodif< kv::interval<double> >::split(dy2, y2, Dfx);
      Ax = prod(trans(Dfx),Y)+prod(Y,Dfx);
      // std::cout << "Ax: " << Ax << std::endl;
      //------------------------- A(x) by using autodiff -----------------------
      kv::veig(Ax,lambda);
      // std::cout << "lambda_AX: " << lambda << std::endl;
      using std::abs;
      cn_tmp = abs(lambda(0));
      for (int i = 1; i < n; i++) {
        if (cn_tmp > abs(lambda(i))) {
          cn_tmp = abs(lambda(i));
        }
      }
      // std::cout << "c_n: " << cn << std::endl;
      // if (in(0,cn_tmp)) {
      if (in(0,cn_tmp)||rad(cn_tmp)>1e-2) { // Uncomment if you want the tight "cn".
        if (eps_tmp==1e-10) {
          std::cout << "Failed in constructing the Lyapunov function" << '\n';
          return 1;
        }
        break;
      }
      y1 = y1_tmp;
      cn = cn_tmp;
      eps = eps_tmp;
      eps_tmp *= 1.1;
    }
    std::cout << "The Lyapunov domain is validated." << '\n';
    std::cout << "c_n: " << cn << std::endl;
    std::cout << "eps = " << eps << '\n';


    // 3. Validate that x(tau_N) is in the Lyapunov domain
    t_n  = 0; // Initialize the blow-up time
    x = x0;
    start = 0;
    end = 400;
    // std::cout << "t: " << t_n << std::endl;
    // std::cout << x << std::endl;
    int r = kv::odelong_maffine(
      Func(),
      x,
      start,
      end,
      // (kv::interval<double>)eps,
      // p,
      // Y,
      kv::ode_param<double>().set_verbose(0).set_order(24).set_restart_max(10).set_epsilon(pow(2,-68)),
      kv::ode_callback_sample<double>(t_n,eps)
    );
    // int r = kv::odelong_maffine0(Func(), x, start, end, kv::ode_param<double>().set_verbose(0).set_order(24).set_restart_max(1),kv::ode_callback_sample<double>(t_n));
    if (r == 0) {
        std::cout << "Cannot calculate solution.\n";
        return 1;
    } else if (r == 1 || r == 3) {
        std::cout << "Solution calculated until t = " << end << ".\n";
        // std::cout << x << "\n";
    } else {
        // std::cout << "Solution completely calculated until t = " << end << ".\n";
        // std::cout << "x_n: " << x << "\n";
        std::cout << "Cannot valitate that x(tau_N) is in the Lyapunov domain." << '\n';
        return 1;
    }
    // std::cout << "Norm(x)" << norm_2(x) << std::endl;
    std::cout << "T_n: " << t_n << std::endl;
    xp = x-p;
    Lx = inner_prod(prod(Y,xp),xp);
    std::cout << "L(x): " << Lx << std::endl;

    // 4. Compute the validated blow-up time of the ODE
    //----------------- This part is specialized for 1-p(x)^4. -----------------
    // kappa = max{1, 6|x^*_2|^2}
    if (1.<6*pow(abs(p(0)),2)) {
      kappa = 6*pow(abs(p(0)),2);
    } else {
      kappa = 1.;
    }
    std::cout << "T_max-T_n: " << mag((
      4*pow((pow(p(1),2)+4*pow(p(0),6))*Lx/c1,0.5)
      +kappa*Lx
      +8./3*abs(p(0))*pow(c1,0.5)*pow(Lx,1.5)
      +0.5*c1*pow(Lx,2))/cn) << std::endl;
    Tmax = kv::interval<double>::hull(t_n,t_n+mag((
      4*pow((pow(p(1),2)+4*pow(p(0),6))*Lx/c1,0.5)
      +kappa*Lx
      +8./3*abs(p(0))*pow(c1,0.5)*pow(Lx,1.5)
      +0.5*c1*pow(Lx,2))/cn));
    //--------------------------------------------------------------------------
    std::cout << "Tmax: " << Tmax << std::endl;
}
