/*
 * ex3_1_dir.cc
 *
 * Validation code of blow-up solution of KS system
 *
 * Space compactification: Directional compactification at r = 0
 * **Notice** this code is specialized for the shockwave (not automatic verification).
 *
 * written  Jul. 12, 2017 Akitoshi Takayasu (with kv-0.4.41)
 *
 */
#include <kv/ode-maffine.hpp>
#include <kv/kraw-approx.hpp>
// #include <kv/eig.hpp>
#include "eig.hpp"
namespace ub = boost::numeric::ublas;
const double L = 1.0;
const int d = 3;
const int N = 8;
const double h = L/N;
ub::vector< kv::interval<double> > r(N-1); // r(0)=r_{1+1/2},...,r(N-2)=r_{N-1+1/2}: control volume
ub::vector< kv::interval<double> > r_pt(N); // r_pt(0)=r_1,...,r_pt(N-1)=r_N: control points
ub::matrix< kv::interval<double> > Y;
ub::vector< kv::interval<double> > x_ast;

// Desingularized vector field of KS
struct Func_orig {
  template <class T> ub::vector<T> operator() (const ub::vector<T>& y, const T& t) {
    // dy(0)=lambda, dy(1)=x_2, dy(2)=x_3,...,dy(N-1)=x_N,
    // dy(N)=y_1, dy(N+1)=y_2,...,dy(2*N-1)=y_N,
    ub::vector<T> dy(2*N);
    T tmp = pow(r_pt(0),1-d)*pow(r(0),d-1)/h/h*(y(0)*(y(1)-1.)-(y(N+1)-y(N)));
    // lambda
    dy(0) = -tmp*y(0)/2;
    // x_i, i=2,...,N
    dy(1) = -tmp*y(1) + pow(r_pt(1),1-d)/h/h*(
      pow(r(1),d-1)*(y(0)*(y(2)-y(1))-(y(N+2)-y(N+1))*y(1))
      -pow(r(0),d-1)*(y(0)*(y(1)-1.)-(y(N+1)-y(N)))
    ); //x_2
    for (int i = 2; i <= N-2; i++) {
      dy(i) = -tmp*y(i) + pow(r_pt(i),1-d)/h/h*(
        pow(r(i),d-1)*(y(0)*(y(i+1)-y(i))-(y(i+N+1)-y(i+N))*y(i))
        - pow(r(i-1),d-1)*(y(0)*(y(i)-y(i-1))-(y(i+N)-y(i+N-1))*y(i-1))
      );
    }
    dy(N-1) = -tmp*y(N-1) - pow(r_pt(N-1),1-d)/h/h*pow(r(N-2),d-1)*(y(0)*(y(N-1)-y(N-2))-(y(2*N-1)-y(2*N-2))*y(N-2));
    // y_i, i=1,...,N
    dy(N) = -tmp*y(N)/2 + y(0)*pow(r_pt(0),1-d)/h/h*pow(r(0),d-1)*(y(N+1)-y(N)) - y(0)*y(N) + 1.;
    for (int i = 1; i < N-1 ; i++) {
      dy(i+N) = -tmp*y(i+N)/2 + y(0)*pow(r_pt(i),1-d)/h/h*(
        pow(r(i),d-1)*(y(i+N+1)-y(i+N))-pow(r(i-1),d-1)*(y(i+N)-y(i+N-1))
      ) - y(0)*y(i+N) + y(i);
    }
    dy(2*N-1) = -tmp*y(2*N-1)/2 - y(0)*pow(r_pt(N-1),1-d)/h/h*pow(r(N-2),d-1)*(y(2*N-1)-y(2*N-2)) - y(0)*y(2*N-1) + y(N-1);
    return dy;
  }
};

// Desingularized vector field (dominant term) of KS
struct Func {
  template <class T> ub::vector<T> operator() (const ub::vector<T>& y, const T& t) {
    // // dy(0)=lambda, dy(1)=x_2, dy(2)=x_3,...,dy(N-1)=x_N,
    // // dy(N)=y_1, dy(N+1)=y_2,...,dy(2*N-1)=y_N,
    ub::vector<T> dy(2*N-1);
    T tmp = pow(r_pt(0),1-d)*pow(r(0),d-1)/h/h*(y(N)-y(N-1));
    // x_i, i=2,...,N
    dy(0) = tmp*y(0) + pow(r_pt(1),1-d)/h/h*(-pow(r(1),d-1)*(y(N+1)-y(N))*y(0)+pow(r(0),d-1)*(y(N)-y(N-1))); //x_2
    for (int i = 1; i <= N-3; i++) {
      dy(i) = tmp*y(i) + pow(r_pt(i+1),1-d)/h/h*(-pow(r(i+1),d-1)*(y(i+N+1)-y(i+N))*y(i)+pow(r(i),d-1)*(y(i+N)-y(i+N-1))*y(i-1));
    }
    dy(N-2) = tmp*y(N-2) + pow(r_pt(N-1),1-d)/h/h*pow(r(N-2),d-1)*(y(2*N-2)-y(2*N-3))*y(N-3);
    // y_i, i=1,...,N
    dy(N-1) = tmp*y(N-1)/2 + 1.;
    for (int i = 1; i < N ; i++) {
      dy(i+N-1) = tmp*y(i+N-1)/2 + y(i-1);
    }
    return dy;
  }
};
// For kraw-approx
struct Func_krawapp {
  template <class T> ub::vector<T> operator() (const ub::vector<T>& y) {
    Func f;
    return f(y, (T) 0);
  }
};
// ode_callback function:
// Integrate the time variable (t) and check the validated solution is in the Lyapunov domain.
namespace kv{
template <class T> struct ode_callback_sample : ode_callback<T> {
    interval<T> &result2;
    T eps;

    ode_callback_sample(interval<T>& result2, T eps) : result2(result2), eps(eps) {}

    virtual bool operator()(const interval<T>& start, const interval<T>& end, const ub::vector< interval<T> >& x_s, const ub::vector< interval<T> >& x_e, const ub::vector< psa< interval<T> > >& result) const {
      result2 += eval(integrate(result(0)),end-start);
      // Stopping criterion
      ub::vector< interval<T> > xp;
      kv::interval<T> Lx;
      xp = x_e-x_ast;
      Lx = inner_prod(prod(Y,xp),xp);
      if (Lx < pow(eps,(kv::interval<T>)2)) return false;
      return true;
    }
};
};

int main()
{
  int n = 2*N;    // # of unknowns
  // Control volumes & control points
  for (int i = 0; i < N-1; i++) {
    r(i) = i+1.;
    r_pt(i) = i+.5;
  }
  r_pt(N-1) = N-.5;
  kv::interval<double> start, end, c1, cn_tmp, cn, t_n, Lx, Tmax;//, kappa;
  ub::vector< kv::interval<double> > x, x0, p, xp, u0;//, v0;
  std::cout.precision(17);
  // x0: initial value on the desingularized ode
  x0.resize(n-1);
  u0.resize(N);
  for (int i = 0; i < N; i++) {
    // u0(i) = 10*exp(-2*pow(r_pt(i)*h,2)) + 20*exp(-2*pow(r_pt(i)*h-0.3,2));
    u0(i) = 100*(1+cos(kv::constants< kv::interval<double> >::pi()*r_pt(i)*h));
  }
  // v0.resize(N);
  // for (int i = 0; i < N; i++) {
  //   v0(i) =
  // }
  for (int i = 0; i < N-1; i++) {
    x0(i) = u0(i+1)/u0(0);
  }
  // std::cout << "x0:" << x0 << '\n';
  // for (int i = N-1; i < n-1; i++) {
  //   x0(i) = u0(i+1)/sqrt(u0(0));
  // }

  // Validation of the blow-up solution
  // 1. Solve the desingularized ode (dominant term) until some time
  //    and determine the critical point at inifity:
  // start = 0;
  // end = 2;
  // x = x0;
  // kv::odelong_maffine(Func(), x, start, end, kv::ode_param<double>().set_verbose(0));
  x.resize(n-1);
  x(0) = -0.036653902557231736;
  x(1) = -8.2755620676517361e-05;
  x(2) = -5.2002967703335875e-10;
  x(N-1) = 0.12277024415402398;
  x(N) = -0.0045000085661491474;
  x(N+1) = -1.0159927755573838e-05;
  // return 0;
  bool b = kv::krawczyk_approx(Func_krawapp(), mid(x), p, 15, 0);
  // print the critical point at infinity
  if (b) {
    std::cout << "x_* = {" << '\n';
    for (int i = 0; i < N-1; i++) {
      std::cout << p(i) << '\n';
    }
    std::cout << "}" << '\n';
    std::cout << "y_* = {" << '\n';
    for (int i = 0; i < N; i++) {
      std::cout << p(i+N-1) << '\n';
    }
    std::cout << "}" << '\n';
  } else {
    std::cout << "Cannot find equilibrium" << '\n';
    return 1;
  }
  // 2. Search the maximum range of the Lyapunov domain:
  // Calculate C1
  Func_orig f; //desingularized vector field of KS
  ub::vector< kv::interval<double> > y1, y2;
  ub::matrix< kv::interval<double> > Dfx;
  ub::vector< kv::autodif< kv::interval<double> > > dy1, dy2;
  // Initial vector of auto diff.
  y1.resize(n);
  for (int i = 1; i < n; i++) {
    y1(i) = p(i-1);
  }
  y1(0) = 0.0;
  x_ast = y1;
  dy1 = kv::autodif< kv::interval<double> >::init(y1);
  dy2 = f(dy1,(kv::autodif< kv::interval<double> >) 0); // f(dy1) = [f(y1), df/dx[y1](Jacobian) (=Dfx)]
  kv::autodif< kv::interval<double> >::split(dy2, y2, Dfx);
  // std::cout << "Df[p]: " << mid(Dfx) << std::endl;
  // Compute eigenvalue of Df(p)
  // Y=Q^T*Q s.t. Q^T*Dfx*Q = T (Shur decomposition)
  ub::vector< kv::complex< kv::interval<double> > > lambda;
  ub::matrix< kv::complex<double> > V, D;
  kv::eig(mid(Dfx),V,D);
  std::cout << "approximate eigenvalue:" << '\n';
  for (int i = 0; i < n; i++) {
    std::cout << D(i,i) << '\n';
    if (D(i,i).real()>0) {
      return 1;
    }
  }
  // Shur decomposition is used for constructing Lyapunov function
  // ub::matrix<double> Q, H;
  // Q.resize(n,n);
  // H.resize(n,n);
  // kv::hess(mid(Dfx),Q,H);
  // kv::francisQR(Q,H);
  // Y = prod(trans(Q),Q);
  // Y = prod(Q,trans(Q));
  Y = ub::identity_matrix<double>(n);
  // std::cout << "Y: " << Y << std::endl;
  kv::veig(Y,lambda);
  // std::cout << "lambda_Y: " << lambda <<  std::endl;
  using std::abs;
  c1 = abs(lambda(0).real());
  for (int i = 1; i < n; i++) {
    if (c1 > abs(lambda(i).real())) {
      c1 = abs(lambda(i).real());
    }
  }
  c1 = 1./c1;
  std::cout << "c_1: " << c1 << std::endl;
  // Calculate Cn by the delta inflation scheme
  double eps_tmp = 1e-10, eps;// Radius of N
  // Func_krawapp fpd;
  ub::vector< kv::interval<double> > y1_tmp,radius;
  ub::matrix< kv::interval<double> > Ax, R;
  y1_tmp.resize(n);
  radius.resize(n);
  bool breakflg=false;
  while (true) {
    //------------------------- A(x) by using autodiff -----------------------
    for (int i = 0; i < n; i++) {
      radius(i) = eps_tmp*kv::interval<double>(-1.,1.);
    }
    y1_tmp = y1 + radius*sqrt(c1);
    // std::cout << "y1 = " << y1 << std::endl;
    dy1 = kv::autodif< kv::interval<double> >::init(y1_tmp);
    dy2 = f(dy1,(kv::autodif< kv::interval<double> >) 0); // f(dy1) = [f(y1), df/dx[y1](Jacobian) (=Dfx)]
    kv::autodif< kv::interval<double> >::split(dy2, y2, Dfx);
// // ///////// Use higher derivatives for enclosing the Jacobian matrix /////////////
//     ub::vector< kv::autodif< kv::autodif< kv::interval<double> > > > ddy1, ddy2;
//     ub::matrix< kv::autodif< kv::interval<double> > > DF_I;
//     ddy1 = kv::autodif< kv::autodif< kv::interval<double> > >::init(dy1);
//     ddy2 = f(ddy1,(kv::autodif< kv::autodif< kv::interval<double> > >) 0);
//     kv::autodif< kv::autodif< kv::interval<double> > >::split(ddy2, dy2, DF_I);
//     // Autodiff at p: use fpd
//     ub::matrix< kv::interval<double> > Dpd;
//     ub::vector< kv::autodif< kv::interval<double> > > dy1_p, dy2_p;
//     dy1_p = kv::autodif< kv::interval<double> >::init(y1);
//     dy2_p = f(dy1_p,(kv::autodif< kv::interval<double> >) 0); // f(dy1) = [f(y1), df/dx[y1](Jacobian) (=Dfx)]
//     dy2 = dy2_p + prod(DF_I,dy1-dy1_p);
//     kv::autodif< kv::interval<double> >::split(dy2, y2, Dfx);
//     // std::cout << "Dpd(p): " << Dpd << std::endl;
//     // Autodiff around the neighborhood of p: use f
//     // std::cout << "Df(I): " << Dfx << std::endl;
// // ////////////
    Ax = prod(trans(Dfx),Y)+prod(Y,Dfx);
    // R = trans(Dfx) + Dfx;
    // R = prod(trans(Q),R);
    // Ax = prod(R,Q);
    // std::cout << "Ax: " << Ax << std::endl;
    //------------------------- A(x) by using autodiff -----------------------
    kv::veig(Ax,lambda);
    // std::cout << "lambda: " << '\n';
    for (int i = 0; i < n; i++) {
      // std::cout << lambda(i).real() << '\n';
      if (lambda(i).real()>0) {
        // std::cout << "A(x) is not negative definite" << '\n';
        breakflg = true;
        break;
      }
    }
    // std::cout << "lambda_AX: " << lambda << std::endl;
    using std::abs;
    cn_tmp = abs(lambda(0));
    for (int i = 1; i < n; i++) {
      if (cn_tmp > abs(lambda(i))) {
        cn_tmp = abs(lambda(i));
      }
    }
    // std::cout << "c_n: " << cn << std::endl;
    // if (in(0,cn_tmp)) {
    if (in(0,cn_tmp)||rad(cn_tmp)>1e-1) { // Uncomment if you want the tight "cn".
      // return 0;
      breakflg = true;
    }
    if (breakflg) {
      if (eps_tmp==1e-10) {
        std::cout << "Failed in constructing the Lyapunov function" << '\n';
        return 1;
      }
      break;
    }
    y1 = y1_tmp;
    cn = cn_tmp;
    eps = eps_tmp;
    eps_tmp *= 1.1;
  }
  std::cout << "The Lyapunov domain is validated." << '\n';
  std::cout << "c_n: " << cn << '\n';
  std::cout << "eps = " << eps << '\n';

  // 3. Validate x(tau_N) is in the Lyapunov domain
  t_n  = 0; // Initialize the blow-up time
  // x0: initial value on the desingularized ode
  x.resize(n);
  x(0) = 1/sqrt(u0(0));
  for (int i = 1; i < n; i++) {
    x(i) = x0(i-1);
  }
  start = 0;
  end = 100;
  int r = kv::odelong_maffine(
    Func_orig(),
    x,
    start,
    end,
    // (kv::interval<double>)eps,
    // p,
    // Y,
    kv::ode_param<double>().set_verbose(0).set_order(24).set_restart_max(10).set_epsilon(pow(2,-68)),
    kv::ode_callback_sample<double>(t_n,eps)
  );
  // int r = kv::odelong_maffine0(Func(), x, start, end, kv::ode_param<double>().set_verbose(0).set_order(24).set_restart_max(1),kv::ode_callback_sample<double>(t_n));
  if (r == 0) {
      std::cout << "Cannot calculate solution.\n";
      return 1;
  } else if (r == 1) {
    std::cout << "Solution incomplitely calculated until t = " << end << ".\n";
    std::cout << x << "\n";
  } else if (r == 3) {
      std::cout << "Solution calculated until t = " << end << ".\n";
      std::cout << x << "\n";
  } else {
      // std::cout << "Solution completely calculated until t = " << end << ".\n";
      // std::cout << "x_n: " << x << "\n";
      std::cout << "Cannot valitate x(tau_N) is in the Lyapunov domain." << '\n';
      return 1;
  }
  // std::cout << "Norm(x)" << norm_2(x) << std::endl;
  std::cout << "T_n: " << t_n << std::endl;
  xp = x-x_ast;
  Lx = inner_prod(prod(Y,xp),xp);
  std::cout << "L(x): " << Lx << std::endl;

  // 4. Compute the validated blow-up time of the ODE
  std::cout << "T_max-T_n: " << mag(2/cn*sqrt(Lx/c1)) << std::endl;
  Tmax = kv::interval<double>::hull(t_n,t_n+mag(2/cn*sqrt(Lx/c1)));
  //--------------------------------------------------------------------------
  std::cout << "Tmax: " << Tmax << std::endl;
}
